import os
from typing import List


class Launcher:
    def __init__(self):

        #: The hashbang line which sets the shell for the PBS script.
        #: If unset, the default is ``#!/usr/bin/env bash``.
        self.hashbang: str = '#!/usr/bin/env bash'

        # these are properties that users typically don't need to set, but are
        # specific to each queueing software
        self.workdir_env_variable: str = ''
        self.profile_filename: str = ''
        self.batch_file_extension: str = ''

    @property
    def profile_filename(self):
        """
        The file to source at the start of the pbs script to set the environment.
        Typical names include '~/.profile', '~/.bashrc', and '~/.cshrc'.
        The default profile filename is '~/.bashrc'

        :type: str
        """
        return self._profile_filename

    @profile_filename.setter
    def profile_filename(self, profile_filename):
        if os.path.isfile(os.path.expanduser(profile_filename)):
            self._profile_filename = profile_filename
        else:
            raise FileNotFoundError('Unable to set profile file.')

    def create_mpi_command(self, command: str,
                           output_root_name: str,
                           openmp_threads: int = None) -> str:
        raise NotImplementedError('Launcher must implement a create_mpi_command method')

    def launch(self, job_name: str, job_body: List[str],
               blocking: bool = True, dependency: str = None) -> str:
        """
        Create a job script and launch the job

        Parameters
        ----------
        job_name:
            The name of the job.
        job_body:
            List of commands to run in the body of the job.
        blocking:
            If true, this function will wait for the job to complete before returning.
            If false, this function will launch the job but not wait for it to finish.
        dependency:
            Jobs that this one depends one. For PBS, these are colon separated in the string

        Returns
        -------
        command_output: str
            The stdout of the launch command. If the job is successfully launch,
            this will be the job id.
        """
        filename = f'{job_name}.{self.batch_file_extension}'
        self.write_job_file(filename, job_name, job_body, dependency)
        return self._run_job(filename, blocking)

    def write_job_file(self, job_filename: str, job_name: str,
                       job_body: List[str], dependency: str = None):
        """
        Create a launch script file in the current directory for the commands defined in ``job_body``.

        Parameters
        ----------
        job_filename:
            name of file to write to
        job_name:
            The name of the job.
        job_body:
            List of commands to run in the body of the job.
        dependency:
            Jobs that this one depends one. For PBS, these are colon separated in the string
        """
        with open(job_filename, mode='w') as fh:
            header = self._create_header(job_name, dependency)
            for line in header:
                fh.write(line + '\n')

            for _ in range(2):
                fh.write('\n')

            fh.write(f'cd {self.workdir_env_variable}\n')
            fh.write('source %s\n' % self.profile_filename)

            for _ in range(1):
                fh.write('\n')

            for line in job_body:
                fh.write(line + '\n')

    def _create_header(self, job_name: str, dependency: str = None) -> List[str]:
        header = self._create_list_of_standard_header_options(job_name)
        header.extend(self._create_list_of_optional_header_lines(dependency))
        return header

    def _create_list_of_standard_header_options(self, job_name: str) -> List[str]:
        return ['']

    def _create_list_of_optional_header_lines(self, dependency: str) -> List[str]:
        return ['']

    def _run_job(self, job_filename: str, blocking: bool, print_command_output: bool = True) -> str:
        raise NotImplementedError('Launcher must implement a _run_job method')



#!/usr/bin/env python
"""
A script to print the directory that a job is running in.
"""
import argparse
from pbs4py.job import PBSJob


def arg_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('job_id', help='The ID number of the job')
    return parser


def main():
    parser = arg_parser()

    args = parser.parse_args()
    job = PBSJob(args.job_id)
    print(job.workdir)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
import os
from typing import List, Union
import numpy as np

from pbs4py.launcher_base import Launcher


class PBS(Launcher):
    def __init__(self, queue_name: str = 'K4-route',
                 ncpus_per_node: int = 40,
                 queue_node_limit: int = 10,
                 time: int = 72,
                 profile_file: str = '~/.bashrc'):
        """
        | A class for creating and running pbs jobs. Default queue properties are for K4.
        | Defaults not set during instantiation can be adjusted by directly modifying attributes.

        Parameters
        ----------
        queue_name:
            Queue name which goes on the "#PBS -N {name}" line of the pbs header
        ncpus_per_node:
            Number of CPU cores per node
        queue_node_limit:
            Maximum number of nodes allowed in this queue
        time:
            The requested job walltime in hours
        profile_file:
            The file setting the environment to source inside the PBS job
        """

        #: The name of the queue which goes on the ``#PBS -N {queue_name}``
        #: line of the pbs header
        self.queue_name: str = queue_name

        #: The number of CPU cores per node.
        self.ncpus_per_node: int = ncpus_per_node

        #: The maximum number nodes allowed by the queue
        self.queue_node_limit: int = queue_node_limit

        #: The requested wall time for the pbs job(s) in hours
        self.time: int = time

        #: The processor model if it needs to be specified.
        #: The associated PBS header line is ``#PBS -l select=#:ncpus=#:mpiprocs=#:model={model}``
        #: If left as `None`, the ``:model={mode}`` will not be added to the header line
        self.model: Union[str, None] = None

        #: The group for the group_list entry of the pbs header if necessary.
        #: The associated PBS header line is ``#PBS -W group_list={group_list}``
        self.group_list: Union[str, None] = None

        #: Requested memory size on the select line. Need to include units in the str.
        #: The associated PBS header line is ``#PBS -l select=#:mem={mem}``
        self.mem: Union[str, None] = None

        #: Index range for PBS array of jobs
        #: The associated PBS header line is ``#PBS -J {array_range}``
        self.array_range: Union[str, None] = None

        #: The mpi execution command name: mpiexec, mpirun, mpiexec_mpt, etc.
        self.mpiexec: str = 'mpiexec'

        #: ``pbs -m`` mail options. 'e' at exit, 'b' at beginning, 'a' at abort
        self.mail_options: str = None

        #: ``pbs -M`` mail list. Who to email when mail_options are triggered
        self.mail_list: Union[str, None] = None

        #: Type of dependency if dependency active.
        #: Default is 'afterok' which only launches the new job if the previous one was successful.
        self.dependency_type: str = 'afterok'

        self.profile_filename: str = profile_file
        self.requested_number_of_nodes: int = 1

        self.workdir_env_variable = '$PBS_O_WORKDIR'
        self.batch_file_extension = 'pbs'
        self.hashbang: str = '#!/usr/bin/env bash'

    @property
    def requested_number_of_nodes(self):
        """
        The number of nodes to request. That is, the 'select' number in the
        ``#PBS -l select={requested_number_of_nodes}:ncpus=40:mpiprocs=40``.

        :type: int
        """
        return self._requested_number_of_nodes

    @requested_number_of_nodes.setter
    def requested_number_of_nodes(self, number_of_nodes):
        self._requested_number_of_nodes = np.min((number_of_nodes, self.queue_node_limit))

    def create_mpi_command(self, command: str,
                           output_root_name: str,
                           openmp_threads: int = None) -> str:
        """
        Wrap a command with mpiexec and route its standard and error output to a file

        Parameters
        ----------
        command:
            The command thats needs to run in parallel
        output_root_name:
            The root name of the output file, {output_root_name}.out.
        openmp_threads:
            The number of openmp threads per mpi process.

        Returns
        -------
        full_command: str
            The full command string.
        """
        omp_env_vars = self._determine_omp_settings(openmp_threads)
        proc_info = self._set_proc_and_omplace_info(openmp_threads)

        direct_output = f'> {output_root_name}.out 2>&1'
        full_command = [omp_env_vars, self.mpiexec, proc_info, command, direct_output]
        return self._filter_empty_strings_from_list_and_combine(full_command)

    def _use_omplace_command(self) -> bool:
        return self.mpiexec == 'mpiexec_mpt'

    def _use_openmp(self, openmp_threads: int or None):
        if type(openmp_threads) == int:
            if openmp_threads > 1:
                return True
        return False

    def _determine_omp_settings(self, openmp_threads: int) -> str:
        if openmp_threads is None:
            return ''

        omp_env_vars = [f'OMP_NUM_THREADS={openmp_threads}']
        if not self._use_omplace_command():
            omp_env_vars.extend(['OMP_PLACES=cores', 'OMP_PROC_BIND=close'])
        return self._filter_empty_strings_from_list_and_combine(omp_env_vars)

    def _filter_empty_strings_from_list_and_combine(self, lis: List[str]) -> str:
        filtered_for_empty_strings = filter(None, lis)
        return ' '.join(filtered_for_empty_strings)

    def _set_proc_and_omplace_info(self, openmp_threads: int) -> str:
        if not self._use_openmp(openmp_threads):
            return ''

        total_num_mpi_procs = (self.ncpus_per_node //
                               openmp_threads) * self.requested_number_of_nodes

        proc_info = f'-np {total_num_mpi_procs}'
        if self._use_omplace_command():
            proc_num_list = ','.join([str(i) for i in range(self.ncpus_per_node)])
            proc_info += f' omplace -c "{proc_num_list}" -nt {openmp_threads} -vv'
        return proc_info

    def _create_list_of_standard_header_options(self, job_name: str) -> List[str]:
        header_lines = [self.hashbang,
                        self._create_job_line_of_header(job_name),
                        self._create_queue_line_of_header(),
                        self._create_select_line_of_header(),
                        self._create_walltime_line_of_header(),
                        self._create_log_name_line_of_header(job_name),
                        self._create_header_line_to_join_standard_and_error_output(),
                        self._create_header_line_to_set_that_job_is_not_rerunnable()]
        return header_lines

    def _create_job_line_of_header(self, job_name: str) -> str:
        return f'#PBS -N {job_name}'

    def _create_queue_line_of_header(self) -> str:
        return f'#PBS -q {self.queue_name}'

    def _create_select_line_of_header(self) -> str:
        select = f'select={self.requested_number_of_nodes}'
        ncpus = f'ncpus={self.ncpus_per_node}'
        mpiprocs = f'mpiprocs={self.ncpus_per_node}'

        select_line = f'#PBS -l {select}:{ncpus}:{mpiprocs}'
        if self.mem is not None:
            select_line += f':mem={self.mem}'
        if self.model is not None:
            select_line += f':model={self.model}'
        return select_line

    def _create_walltime_line_of_header(self) -> str:
        return f'#PBS -l walltime={self.time}:00:00'

    def _create_log_name_line_of_header(self, job_name: str) -> str:
        return f'#PBS -o {job_name}_pbs.log'

    def _create_header_line_to_join_standard_and_error_output(self):
        return '#PBS -j oe'

    def _create_header_line_to_set_that_job_is_not_rerunnable(self) -> str:
        return '#PBS -r n'

    def _create_list_of_optional_header_lines(self, dependency):
        header_lines = []
        header_lines.extend(self._create_group_list_header_line())
        header_lines.extend(self._create_array_range_header_line())
        header_lines.extend(self._create_mail_options_header_lines())
        header_lines.extend(self._create_job_dependencies_header_line(dependency))
        return header_lines

    def _create_group_list_header_line(self) -> List[str]:
        if self.group_list is not None:
            return [f'#PBS -W group_list={self.group_list}']
        else:
            return []

    def _create_array_range_header_line(self) -> List[str]:
        if self.array_range is not None:
            return [f'#PBS -J {self.array_range}']
        else:
            return []

    def _create_mail_options_header_lines(self) -> List[str]:
        header_lines = []
        if self.mail_options is not None:
            header_lines.append(f'#PBS -m {self.mail_options}')
        if self.mail_list is not None:
            header_lines.append(f'#PBS -M {self.mail_list}')
        return header_lines

    def _create_job_dependencies_header_line(self, dependency) -> List[str]:
        if dependency is not None:
            return [f'#PBS -W depend={self.dependency_type}:{dependency}']
        else:
            return []

    def _run_job(self, job_filename: str, blocking: bool, print_command_output=True) -> str:
        options = ''
        if blocking:
            options += '-Wblock=true'
        command_output = os.popen(f'qsub {options} {job_filename}').read().strip()
        if print_command_output:
            print(command_output)
        return command_output

    # Alternate constructors for NASA HPC queues
    @classmethod
    def k4(cls, time: int = 72, profile_file: str = '~/.bashrc'):
        """
        Constructor for the K4 queues on LaRC's K cluster including K4-standard-512.

        Parameters
        ----------
        time:
            The requested job walltime in hours
        profile_file:
            The file setting the environment to source inside the PBS job
        """
        return cls(queue_name='K4-route', ncpus_per_node=40,
                   queue_node_limit=16, time=time,
                   profile_file=profile_file)

    @classmethod
    def k3(cls, time: int = 72, profile_file: str = '~/.bashrc'):
        """
        Constructor for the K3 queues on LaRC's K cluster including K3-standard-512.

        Parameters
        ----------
        time:
            The requested job walltime in hours
        profile_file:
            The file setting the environment to source inside the PBS job
        """
        return cls(queue_name='K3-route', ncpus_per_node=16,
                   queue_node_limit=40, time=time,
                   profile_file=profile_file)

    @classmethod
    def k3a(cls, time: int = 72, profile_file: str = '~/.bashrc'):
        """
        Constructor for the K3a queue on LaRC's K cluster.

        Parameters
        ----------
        time:
            The requested job walltime in hours
        profile_file:
            The file setting the environment to source inside the PBS job
        """
        return cls(queue_name='K3a-route', ncpus_per_node=16,
                   queue_node_limit=25, time=time,
                   profile_file=profile_file)

    @classmethod
    def nas(cls, group_list: str, proc_type: str = 'broadwell', queue_name: str = 'long',
            time: int = 72, profile_file: str = '~/.bashrc'):
        """
        Constructor for the queues at NAS. Must specify the group_list

        Parameters
        ----------
        group_list:
            The charge number or group for the group_list entry of the pbs header.
            The associated PBS header line is "#PBS -W group_list={group_list}".
        proc_type:
            The type of processor to submit to. Can write out or just the first 3 letters:
            'cas', 'sky', 'bro', 'has', 'ivy', 'san'.
        queue_name:
            Which queue to submit to: devel, debug, normal, long, etc.
        time:
            The requested job walltime in hours
        profile_file:
            The file setting the environment to source inside the PBS job
        """

        if 'cas' in proc_type.lower():
            ncpus_per_node = 40
            model = 'cas_ait'
        elif 'sky' in proc_type.lower():
            ncpus_per_node = 40
            model = 'sky_ele'
        elif 'bro' in proc_type.lower():
            ncpus_per_node = 28
            model = 'bro'
        elif 'has' in proc_type.lower():
            ncpus_per_node = 24
            model = 'has'
        elif 'ivy' in proc_type.lower():
            ncpus_per_node = 20
            model = 'ivy'
        elif 'san' in proc_type.lower():
            ncpus_per_node = 20
            model = 'san'
        else:
            raise ValueError('Unknown NAS processor selection')

        pbs = cls(queue_name=queue_name, ncpus_per_node=ncpus_per_node,
                  queue_node_limit=int(1e6), time=time, profile_file=profile_file)

        pbs.group_list = group_list
        pbs.model = model
        return pbs


class FakePBS(PBS):
    def __init__(self):
        """
        A fake PBS class for directly running commands while still calling as
        if it were a standard PBS driver.
        This can be used to seemless switch between modes where PBS jobs are
        launched for each "job", or using a FakePBS object when you don't want to
        launch a new pbs job for each "job", e.g., driving a script
        while already within the PBS job.
        """
        super().__init__()

    def launch(self, job_name: str, job_body: List[str],
               blocking: bool = True, dependency: str = None) -> str:
        """
        Runs the commands in the job_body

        Parameters
        ----------
        job_name:
            [ignored]
        job_body:
            List of commands to run
        blocking:
            [ignored]
        dependency:
            [ignored]

        Returns
        -------
        pbs_command_output: str
            Empty string but returning something to match true PBS launch output
        """
        for line in job_body:
            print(line)
            os.system(line)
        return ''



import os
from typing import List

from pbs4py.launcher_base import Launcher


class BSUB(Launcher):
    def __init__(self,
                 project: str,
                 ngpu_per_node: int = 6,
                 time: int = 72,
                 profile_filename: str = '~/.bashrc'):
        """
        A Class for creating and running jobs using the Department of Energy
        batch system.

        Parameters
        ----------
        project:
            The project which to charge for submitted jobs
        ngpu_per_node:
            The number of GPUs per compute node
        time:
            The requested wall time for the job(s) in hours
        profile_filename:
            The file setting the environment to source inside the PBS job
        """
        #: The requested wall time for the job(s) in hours
        self.time: int = time

        #: The number of compute nodes requested
        self.requested_number_of_nodes: int = 1

        #: The number of GPUs per compute node.
        self.ngpu_per_node: int = ngpu_per_node

        #: The project which to charge for submitted jobs
        self.project: str = project

        #: Mail a job report when complete
        self.mail_when_complete: bool = True

        self.profile_filename = profile_filename
        self.workdir_env_variable = '$LS_SUBCWD'
        self.batch_file_extension = 'lsf'

    def create_mpi_command(self, command: str,
                           output_root_name: str,
                           openmp_threads: int = 1) -> str:
        num_mpi_procs = self.requested_number_of_nodes * self.ngpu_per_node
        command = f'jsrun -n {num_mpi_procs} -a 1 -c {openmp_threads} -g 1 {command} > {output_root_name}.out 2>&1'
        return command

    def _create_list_of_standard_header_options(self, job_name: str) -> List[str]:
        header_lines = [self.hashbang,
                        self._create_project_line_of_header(),
                        self._create_job_name_line_of_header(job_name),
                        self._create_number_of_nodes_line_of_header(),
                        self._create_wall_time_line_of_header()]
        return header_lines

    def _create_project_line_of_header(self) -> str:
        return f'#BSUB -P {self.project}'

    def _create_job_name_line_of_header(self, job_name: str) -> str:
        return f'#BSUB -J {job_name}'

    def _create_number_of_nodes_line_of_header(self) -> str:
        return f'#BSUB -nnodes {self.requested_number_of_nodes}'

    def _create_wall_time_line_of_header(self) -> str:
        return f'#BSUB -W {self.time}:00'

    def _create_list_of_optional_header_lines(self, dependency: str) -> List[str]:
        header_lines = []
        header_lines.extend(self._create_job_dependency_header_line(dependency))
        header_lines.extend(self._create_mail_header_line())
        return header_lines

    def _create_job_dependency_header_line(self, dependency: str) -> List[str]:
        if dependency is not None:
            return [f'#BSUB -w ended({dependency})']
        else:
            return []

    def _create_mail_header_line(self) -> List[str]:
        if self.mail_when_complete:
            return [f'#BSUB -N']
        else:
            return []

    def _run_job(self, job_filename: str, blocking: bool, print_command_output: bool = True) -> str:
        if blocking:
            print('Warning: Blocking for bsub not implemented')

        command = f'bsub {job_filename}'
        if print_command_output:
            print(command)
        return os.popen(command).read()

    def _parse_job_id_out_of_bsub_output(self, bsub_output: str) -> int:
        return int(bsub_output.split('>')[0].split('<')[-1])



#!/usr/bin/env python
"""
A script to delete active PBS jobs of the current user.
The list of jobs to be deleted can be filtered by id range, job name substring, and queue.
For safety, the default behavior is to show the user which jobs will be deleted and ask for confirmation
before any jobs are deleted.
"""
import os
import argparse
import re
from typing import List

from pbs4py.job import PBSJob


def arg_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('--id_range',
                        nargs=2,
                        default=(-1, -1),
                        help='Delete jobs in a range of id numbers, [min id, max id]')
    parser.add_argument('--queue',
                        default='',
                        help='Delete jobs in a specific queue')
    parser.add_argument('--name',
                        default='',
                        help='Delete jobs in a specific string in the name')
    parser.add_argument('--confirm',
                        action='store_true',
                        dest='confirm',
                        help='Whether to prompt the user for confirmation of before deleting')
    parser.add_argument('--no-confirm', dest='confirm', action='store_false')
    parser.set_defaults(confirm=True)
    return parser


def get_active_jobs_for_user():
    user_name = os.environ.get('USER')
    qstat_output = os.popen(f'qstat -u {user_name}').read().split('\n')

    # remove header from qstat command
    qstat_output = qstat_output[3:]

    jobs = []
    for line in qstat_output:
        if line:
            id = int(re.match('\s*[0-9]+', line)[0])
            jobs.append(PBSJob(id))
    return jobs


def filter_jobs_to_delete_by_id_range(user_jobs: List[PBSJob], min_id: int, max_id: int):
    return [job for job in user_jobs if (job.id >= min_id and job.id <= max_id)]


def filter_jobs_to_delete_by_queue(user_jobs: List[PBSJob], queue: str):
    return [job for job in user_jobs if job.queue == queue]


def filter_jobs_to_delete_by_name_substring(user_jobs: List[PBSJob], name_substring: str):
    return [job for job in user_jobs if name_substring in job.name]


def delete_jobs(jobs: List[PBSJob]):
    for job in jobs:
        job.qdel(echo_command=True)


def print_jobs_that_will_be_deleted(jobs_to_delete: List[PBSJob]):
    print('Found the following jobs:')
    print('------------------------')
    for job in jobs_to_delete:
        print(f'Job: id = {job.id}, name = {job.name}, queue: {job.queue}')


def user_confirms():
    prompt = 'Delete these jobs? [y/n]'
    valid = {"yes": True, "y": True, "no": False, "n": False}

    while True:
        print(prompt)
        choice = input().lower()
        if choice in valid:
            return valid[choice]
        else:
            print("Please respond with 'yes' or 'no' ", "(or 'y' or 'n').\n")


def main():
    parser = arg_parser()

    args = parser.parse_args()
    confirm = args.confirm
    min_id = int(args.id_range[0])
    max_id = int(args.id_range[1])
    queue = args.queue
    name_substring = args.name

    jobs_to_delete = get_active_jobs_for_user()

    if min_id > 0 and max_id > 0:
        print('Filtering by id range')
        jobs_to_delete = filter_jobs_to_delete_by_id_range(jobs_to_delete, min_id, max_id)
    if queue:
        print('Filtering by queue')
        jobs_to_delete = filter_jobs_to_delete_by_queue(jobs_to_delete, queue)
    if name_substring:
        print('Filtering by name')
        jobs_to_delete = filter_jobs_to_delete_by_name_substring(jobs_to_delete, name_substring)

    if len(jobs_to_delete) == 0:
        print(f'No active jobs found for user with specified filters')
        exit()

    if confirm:
        print_jobs_that_will_be_deleted(jobs_to_delete)
        if user_confirms():
            delete_jobs(jobs_to_delete)
        else:
            print('Skipping')

    else:
        delete_jobs(jobs_to_delete)


if __name__ == '__main__':
    main()



import os
from typing import List, Union


class PBSJob:
    def __init__(self, id: int):
        """
        A class for querying information and managing a particular submitted
        pbs job. For the id number in the constructor, the qstat command will
        be used to populate the attributes of the job.

        Parameters
        ----------
        id:
            The id number of the job
        """

        #: The ID number of the job
        self.id: int = id

        #: The name of the job
        self.name: str = ''

        #: The model attribute on the select line from the job submission
        self.model: str = ''

        #: The queue which this job was submitted to
        self.queue: str = ''

        #: Whether the job is queued, running, or finished
        self.state: str = ''

        #: The value of $PBS_O_WORKDIR
        self.workdir: str = ''

        self.read_properties_from_qstat()

    def read_properties_from_qstat(self):
        """
        Use qstat to get the current attributes of this job
        """
        qstat_output = self._run_qstat_to_get_full_job_attributes()
        if self._is_a_known_job(qstat_output):
            self._parse_attributes_from_qstat_output(qstat_output)
        else:
            self._set_empty_attributes()

    def qdel(self, echo_command: bool = True) -> str:
        """
        Call qdel to delete this job

        Parameters
        ----------
        echo_command:
            Whether to print the command before running it

        Returns
        -------
        command_output: str
            The output of the shell command
        """
        command = f'qdel {self.id}'
        if echo_command:
            print(command)
        return os.popen(command).read()

    def _run_qstat_to_get_full_job_attributes(self) -> Union[List[str], str]:
        return os.popen(f'qstat -xf {self.id}').read().split('\n')

    def _is_a_known_job(self, qstat_output):
        return not 'Unknown Job Id' in qstat_output

    def _parse_attributes_from_qstat_output(self, qstat_output: List[str]):
        qstat_dict = self._convert_qstat_output_to_a_dictionary(qstat_output)

        self.name: str = qstat_dict['Job_Name']
        self.queue: str = qstat_dict['queue']
        self.state: str = qstat_dict['job_state']
        self.workdir = self._parse_workdir(qstat_dict)

        if 'model' in qstat_dict['Resource_List.select']:
            self.model = qstat_dict['Resource_List.select'].split('model=')[-1]
        else:
            self.model = ''

    def _set_empty_attributes(self):
        self.name = ''
        self.model = ''
        self.queue = ''
        self.state = ''
        self.workdir = ''

    def _parse_workdir(self, qstat_dict: dict) -> str:
        return qstat_dict['Variable_List'].split('PBS_O_WORKDIR=')[-1].split(',')[0]

    def _convert_qstat_output_to_a_dictionary(self, qstat_output: List[str]) -> dict:
        qstat_dict = {}
        for line in qstat_output:
            if '=' in line:
                split_line = line.split('=', 1)
                key = split_line[0].strip()
                value = split_line[1].strip()
                qstat_dict[key] = value
        return qstat_dict



"""
Notices:
Copyright 2022 United States Government as represented by the Administrator of
the National Aeronautics and Space Administration. No copyright is claimed in
the United States under Title 17, U.S. Code. All Other Rights Reserved.

Third Party Software:

This software calls the following third party software, which is subject to the
terms and conditions of its licensor, as applicable at the time of licensing.
Third party software is not bundled with this software, but may be available
from the licensor. License hyperlinks are provided here for information purposes
only: numpy, BSD 3-Clause "New" or "Revised" License,
https://github.com/numpy/numpy/blob/main/LICENSE.txt.

Disclaimers
No Warranty: THE SUBJECT SOFTWARE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY OF
ANY KIND, EITHER EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED
TO, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL CONFORM TO SPECIFICATIONS, ANY
IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
FREEDOM FROM INFRINGEMENT, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL BE ERROR
FREE, OR ANY WARRANTY THAT DOCUMENTATION, IF PROVIDED, WILL CONFORM TO THE
SUBJECT SOFTWARE. THIS AGREEMENT DOES NOT, IN ANY MANNER, CONSTITUTE AN
ENDORSEMENT BY GOVERNMENT AGENCY OR ANY PRIOR RECIPIENT OF ANY RESULTS,
RESULTING DESIGNS, HARDWARE, SOFTWARE PRODUCTS OR ANY OTHER APPLICATIONS
RESULTING FROM USE OF THE SUBJECT SOFTWARE.  FURTHER, GOVERNMENT AGENCY
DISCLAIMS ALL WARRANTIES AND LIABILITIES REGARDING THIRD-PARTY SOFTWARE, IF
PRESENT IN THE ORIGINAL SOFTWARE, AND DISTRIBUTES IT "AS IS."

Waiver and Indemnity:  RECIPIENT AGREES TO WAIVE ANY AND ALL CLAIMS AGAINST THE
UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL AS ANY
PRIOR RECIPIENT.  IF RECIPIENT'S USE OF THE SUBJECT SOFTWARE RESULTS IN ANY
LIABILITIES, DEMANDS, DAMAGES, EXPENSES OR LOSSES ARISING FROM SUCH USE,
INCLUDING ANY DAMAGES FROM PRODUCTS BASED ON, OR RESULTING FROM, RECIPIENT'S USE
OF THE SUBJECT SOFTWARE, RECIPIENT SHALL INDEMNIFY AND HOLD HARMLESS THE UNITED
STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL AS ANY PRIOR
RECIPIENT, TO THE EXTENT PERMITTED BY LAW.  RECIPIENT'S SOLE REMEDY FOR ANY SUCH
 MATTER SHALL BE THE IMMEDIATE, UNILATERAL TERMINATION OF THIS AGREEMENT.
"""
import os
from setuptools import setup, find_packages


__package_name__ = "pbs4py"
__package_version__ = "1.0.0"

root =os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(root, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name=__package_name__,
    version=__package_version__,
    description=("PBS scripting utilities"),
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Kevin Jacobson",
    author_email="kevin.e.jacobson@nasa.gov",
    zip_safe=False,
    packages=find_packages(),
    scripts=['pbs4py/scripts/qdel_user_jobs.py',
             'pbs4py/scripts/job_dir.py'],
    install_requires=['numpy'],
    python_requires='>=3.6'
)



import os
import numpy as np
import time
from datetime import datetime
from typing import List
from .pbs import PBS
from .directory_utils import cd


class BatchJob:
    def __init__(self, name: str, body: List[str]):
        """
        Class for individual PBS jobs within a batch of jobs

        Can be used as a context manager to enter/exit a directory
        with the job's name
        """

        #: str: Name of the job.
        self.name = name

        #: List[str]: list of commands to run in PBS job
        self.body = body

        #: str: pbs job identifier returned by qsub
        self.id = None

    def get_pbs_job_state(self) -> str:
        """
        Get the job's status after it has been submitted.
        Returns the entry of job_state in the qstat information, e.g.,
        'Q', 'R', 'F', 'H', etc.

        """
        if self.id is not None:
            return os.popen(f'qstat -xf {self.id} | grep job_state').read().split()[-1]
        else:
            return ''

    def __enter__(self):
        self.savedPath = os.getcwd()
        os.chdir(self.name)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.savedPath)


class PBSBatch:
    def __init__(self, pbs: PBS, jobs: List[BatchJob], use_separate_directories: bool = True):
        """
        Batch of PBS jobs. Assumes all jobs required the same
        job request size. By default, separate directories with the job's name
        will be used to separate output files.

        Parameters
        ----------
        pbs:
            PBS handler that will be used to submit the jobs
        jobs:
            List of Job objects that will be run
        use_separate_directories:
            whether to run each job in a separate directory with the job's name
        """
        self.pbs = pbs
        self.jobs = jobs
        self.use_separate_directories = use_separate_directories

    def create_directories(self):
        """
        Create the set of directories with the jobs' names
        """
        for job in self.jobs:
            if not os.path.exists(job.name):
                os.mkdir(job.name)

    def launch_all_jobs(self, wait_for_jobs_to_finish: bool = False,
                        check_frequency_in_secs: float = 30):
        """
        Launch of the all of the jobs in the list. Stores the pbs
        job id in the job objects

        Parameters
        ----------
        wait_for_jobs_to_finish:
            If True, the jobs will be submitted, and this function will not return
            until all of the jobs are finished.

        check_frequency_in_secs:
            Time interval to wait before checking if all jobs are done. Only relevant
            if ``wait_for_jobs_to_finish`` is True.
        """
        self._launch_jobs_in_a_list(self.jobs)
        if wait_for_jobs_to_finish:
            self.wait_for_all_jobs_to_finish(check_frequency_in_secs=check_frequency_in_secs)

    def launch_jobs_with_limit(self, max_jobs_at_a_time: int = 20,
                               check_frequency_in_secs: float = 30):
        """
        The "courteous" version of launch_all_jobs(wait_for_jobs_to_finish=True) and where a limit
        is set for the maximum number of jobs running or in the queue at a time since some people
        may not like if you submit 1000 jobs at once.

        Parameters
        ----------
        max_jobs_at_a_time:
            Limit for number of jobs to have queued, running, or held at a time

        check_frequency_in_secs:
            Time interval to wait before checking if jobs' statuses.
        """
        total_num_of_jobs = len(self.jobs)

        next_job_to_submit = 0
        while True:
            states = self._get_job_states(self.jobs[:next_job_to_submit])
            num_active_jobs = self._count_number_of_jobs_running_queued_or_held(states)
            if num_active_jobs < max_jobs_at_a_time:
                end_index = np.min(
                    (total_num_of_jobs, next_job_to_submit + max_jobs_at_a_time - num_active_jobs))
                self._launch_jobs_in_a_list(self.jobs[next_job_to_submit:end_index])
                next_job_to_submit = end_index

            states = self._get_job_states(self.jobs[:next_job_to_submit])
            self._print_summary_of_job_states(states)
            if self._all_jobs_submitted(next_job_to_submit):
                if not self._any_jobs_are_still_running_queued_or_held(states):
                    break
            time.sleep(check_frequency_in_secs)

    def wait_for_all_jobs_to_finish(self, check_frequency_in_secs: float = 30):
        """
        A blocking check for all the jobs in the batch to finish. Can be paired with
        ``launch_all_jobs``.

        Parameters
        ----------
        check_frequency_in_secs:
            How often to check and print the jobs' states
        """
        while True:
            states = self._get_job_states(self.jobs)
            self._print_summary_of_job_states(states)
            if self._any_jobs_are_still_running_queued_or_held(states):
                time.sleep(check_frequency_in_secs)
            else:
                break

    def _launch_jobs_in_a_list(self, jobs: List[BatchJob]):
        for job in jobs:
            dirname = job.name if self.use_separate_directories else '.'
            with cd(dirname):
                job.id = self.pbs.launch(job.name, job.body, blocking=False)

    def _all_jobs_submitted(self, next_job_to_submit):
        return next_job_to_submit == len(self.jobs)

    def _get_job_states(self, jobs: List[BatchJob]) -> List[str]:
        states = []
        for job in jobs:
            states.append(job.get_pbs_job_state())
        return states

    def _count_number_of_jobs_running_queued_or_held(self, pbs_states: List[str]):
        return pbs_states.count('R') + pbs_states.count('Q') + pbs_states.count('H')

    def _any_jobs_are_still_running_queued_or_held(self, pbs_states):
        return self._count_number_of_jobs_running_queued_or_held(pbs_states) > 0

    def _print_summary_of_job_states(self, states: List[str]):
        running = states.count('R')
        queued = states.count('Q')
        finished = states.count('F')
        other = len(states) - running - queued - finished
        print(f'Job states at {datetime.now().isoformat()}:')
        print(f'  Queued:        {queued}')
        print(f'  Running:       {running}')
        print(f'  Finished:      {finished}')

        num_of_jobs_not_yet_submitted = len(self.jobs)-len(states)
        if num_of_jobs_not_yet_submitted > 0:
            print(f'  Yet to submit: {num_of_jobs_not_yet_submitted}')
        print(f'  Other:         {other}')
